// ============================================================================
// CBP + FBO = MAS22 Simulation Framework
// 20-12-21
// 28-01-22
// MAS22_SkyMaster.cpp
// ============================================================================

#include "MAS22_SkyMaster.h"
#include "MAS22_Actor.h"
#include "MAS22_Structs.h"
#include "rkf45.hpp"
#include "rhs_SkyMaster.h"
#include "Components/CapsuleComponent.h"
#include "Components/StaticMeshComponent.h"

// Sets default values
AMAS22_SkyMaster::AMAS22_SkyMaster() {
	PrimaryActorTick.bCanEverTick = false;

	CapsuleCollider = CreateDefaultSubobject<UCapsuleComponent>("CapsuleCollider");
	CapsuleCollider->SetupAttachment(GetRootComponent());
	CapsuleCollider->SetCapsuleHalfHeight(200);
	CapsuleCollider->SetCapsuleRadius(100);
	CapsuleCollider->SetRelativeLocation(FVector(0, 0, 0));

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset1(TEXT("/Game/SkyMaster/skymasterbase_1.skymasterbase"));
	UStaticMesh* Asset1 = MeshAsset1.Object;

	Base = CreateDefaultSubobject<UStaticMeshComponent>("Base");
	Base->SetStaticMesh(Asset1);
	Base->SetRelativeLocation(FVector(0, 0, 0));
	Base->SetupAttachment(CapsuleCollider);

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset2(TEXT("/Game/SkyMaster/skymaster3_1.skymaster3"));
	UStaticMesh* Asset2 = MeshAsset2.Object;

	LeftCar = CreateDefaultSubobject<UStaticMeshComponent>("LeftCar");
	LeftCar->SetStaticMesh(Asset2);
	LeftCar->SetRelativeLocation(FVector(100, 0, 250));
	LeftCar->SetupAttachment(CapsuleCollider);

	RightCar = CreateDefaultSubobject<UStaticMeshComponent>("RightCar");
	RightCar->SetStaticMesh(Asset2);
	RightCar->SetRelativeLocation(FVector(-100, 0, 250));
	RightCar->SetupAttachment(CapsuleCollider);

}

void AMAS22_SkyMaster::ActorInitialize() {
	Super::ActorInitialize();

	u    = new double[2];
	dudt = new double[2];
	parameters = new double[4];
	SetInitialConditions();
	InitializeSolver();

}

void AMAS22_SkyMaster::SetInitialConditions() {
	Super::SetInitialConditions();
	// (15) ----------------------------
	u[0] = initTheta;
	u[1] = 0.0;

}

void AMAS22_SkyMaster::InitializeSolver() {

	parameters[0] = mass;
	parameters[1] = armLength;
	parameters[2] = gravity;
	parameters[3] = damping;
	
	 setParams_SM(parameters);  // This is in the RHS

	// RKF Stuff
	flag = 1;
	t = 0.0;
	relerr = 100 * sqrt(r8_epsilon());
	abserr = 100 * sqrt(r8_epsilon());
	GEngine->AddOnScreenDebugMessage(-1, 10.0f, FColor::Yellow,
		FString::Printf(TEXT("MAS22_MonsterTruck:: relerr  = %e"), relerr));

}

void AMAS22_SkyMaster::BuildHUDLines() {
	HUDLine line;
	Super::BuildHUDLines();
	HUDLines.Empty();
	bHUD_exists = true;
	line.message = FString::Printf(TEXT("time = %f"), time);
	line.colour = FColor::Red;
	line.value = 0;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("theta = %f"), u[0]);
	line.colour = FColor::Yellow;
	line.value = 0;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("omega = %f"), u[1]);
	line.colour = FColor::Yellow;
	line.value = 0;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("maxTheta = %f"), maxTheta);
	line.colour = FColor::Yellow;
	line.value = 0;
	HUDLines.Add(line);
}

void AMAS22_SkyMaster::LogData() {
	int m;
	m = (int(logInterval / dT));
	if (itn % m == 0) {
		dataRecords.Add(FDataRecord_SM{ time, u[0], u[1] });
	}

}

void AMAS22_SkyMaster::ComputeDT() {
	if (bRunning) {
		Computation();
		Visualization(dT);
		BuildHUDLines();
		LogData();
		if (u[0] > maxTheta) {
			maxTheta = u[0];
		}
		if (u[0] < 0) {
			maxTheta = 0;
		}

		itn++;
		//GEngine->AddOnScreenDebugMessage(-1, 4.0f, FColor::Yellow,
		//	FString::Printf(TEXT("MAS22_MonsterTruck:: disp = %f"),u[0]));
	}
}

void AMAS22_SkyMaster::Computation() {
	// (21) --------------------
	switch (forcingType) {
	case periodic_SM:
		forcing = forcingAmplitude * FMath::Cos(2 * PI * time / forcingPeriod);
		applyPeriodicDrive(forcing);
		break;
	}

	switch (solver) {
	case SolverEnum::EULER:
		Euler(dT);
		break;
		// (21) -----------------------------
	case SolverEnum::RUNGE_KUTTA_FEHLBERG:
		RKF(dT);
		break;
	default:
		break;
	}

}

void AMAS22_SkyMaster::Euler(double deltaT) {

	dT = deltaT;

    rhs_SkyMaster(time, u, dudt);

	u[0] += dudt[0] * dT;
	u[1] += dudt[1] * dT;
	time += dT;


}

void AMAS22_SkyMaster::RKF(double deltaT) {
	flag = 2;
	tout = time + deltaT;

	flag = r8_rkf45(rhs_SkyMaster, 2, u, dudt, &time, tout, &relerr, abserr, flag);

}



void AMAS22_SkyMaster::Visualization(float deltaTime) {
	// Following removed, placed into header as a private. Static variables seem to be shared between instances.
	//static float totalRot = 0;

	LeftCar->SetRelativeRotation(FRotator(0,0,u[0]*180*scaling/PI).Quaternion());
	RightCar->SetRelativeRotation(FRotator(0,0,-u[0]* 180*scaling / PI).Quaternion());

}


void AMAS22_SkyMaster::EndPlay(const EEndPlayReason::Type EndPlayReason) {
	Super::EndPlay(EndPlayReason);

	// (19) Dump Octave file -------------
	FString logFile = FPaths::ProjectDir();
	FString fileName = "OctaveFiles/";
	fileName.Append(actorName);
	fileName.Append(".m");
	logFile.Append(fileName);

	int nrRecords = dataRecords.Num();
	FString stringToSave = "A=[\n";

	for (int i = 0; i < nrRecords; i++) {
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].time));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].theta));
		stringToSave.Append(FString::Printf(TEXT("%e\n"),  dataRecords[i].omega));
	}

	stringToSave.Append("];\n\n");
	stringToSave.Append("time  = A(:,1);\n");
	stringToSave.Append("theta = 180*A(:,2)/pi;\n");
	stringToSave.Append("omega = 180*A(:,3)/pi;\n");

	stringToSave.Append("figure(1)\n");
	//stringToSave.Append("xlabel('time (secs)')\n");
	//stringToSave.Append("ylabel('theta (degrees)')\n");
	stringToSave.Append("subplot(2,1,1)\n");
	stringToSave.Append("plot(time,theta);\n");
	stringToSave.Append("title('Theta vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('theta (degrees)')\n");

	stringToSave.Append("subplot(2,1,2)\n");
	stringToSave.Append("plot(time,omega);\n");
	stringToSave.Append("title('Omega vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('omega (degree/secs)')\n");

	stringToSave.Append("figure(2)\n");
	stringToSave.Append("plot(theta,omega);\n"); 
	stringToSave.Append("title('Omega vs. Theta Phase Plane')\n");
	stringToSave.Append("xlabel('theta (degrees)')\n");
	stringToSave.Append("ylabel('omega (degree/secs)')\n");

	FFileHelper::SaveStringToFile(stringToSave, *logFile);

}

